'''
    Calculate the completeness of the data contained in an oil record.

    The top-level interface is a single function:
        completeness(pyjson_of_an_oil_record)

    (pyjson is a python dict that is a match for the JSON)

    It performs calculations that were designed by RobertJ, and returns a value
    with a scale of 0->100%.
'''
import logging

from .oil import Oil

# from ..common.measurement import MassFraction, Temperature

logger = logging.getLogger(__name__)


# is this function needed???
# answer: SRP, it's the right thing to do
def set_completeness(oil):
    oil.metadata.model_completeness = completeness(oil)


def completeness(oil):
    '''
        Calculate the completeness of the data contained in an oil record.

        :param oil: The oil record to be validated, in json-compatible python
                    data structure.
    '''
    res = 0
    for check_func in CHECKS:
        res += check_func(oil)

    return round(res * 10.0)


def check_emulsion_water_content(oil):
    '''
        One emulsion water content in any subsample. Score = 2.5
    '''
    sub_samples = oil.sub_samples

    for sample in sub_samples:
        emuls = sample.environmental_behavior.emulsions
        for e in emuls:
            if (is_measurement_good(e.water_content)):
                return 2.5  # only need one valid emulsion water content

    return 0.0


def check_density(oil):
    '''
        Fresh oil: One density or API. Score = 1
    '''

    if oil.metadata.API is not None:
        return 1.0

    if len(oil.sub_samples) > 0:
        ss = oil.sub_samples[0]

        densities = ss.physical_properties.densities

        for d in densities:
            if (is_measurement_good(d.density) and
                    is_measurement_good(d.ref_temp)):
                return 1.0

    return 0.0


def check_second_density(oil):
    '''
    Fresh oil: Second density separated by temperature.

    Score = deltaT/40 but not greater than 0.5

    maxDeltaT: The difference between the lowest and highest
    measurement in the set.
    '''
    if len(oil.sub_samples) > 0:
        ss = oil.sub_samples[0]
        densities = ss.physical_properties.densities

        temps = [d.ref_temp.converted_to('C').value
                 for d in densities
                 if d.ref_temp is not None]

        if len(temps) >= 2:
            t1, *_, t2 = sorted([t for t in temps if t is not None])
            delta_t = t2 - t1

            if delta_t > 0.0:
                return min(delta_t / 40.0, 0.5)

    return 0.0


def check_viscosity(oil):
    '''
    Fresh oil: One viscosity. Score = 0.5
    '''
    if len(oil.sub_samples) > 0:
        ss = oil.sub_samples[0]
        kvis = ss.physical_properties.kinematic_viscosities
        dvis = ss.physical_properties.dynamic_viscosities

        for v in kvis:
            if (is_measurement_good(v.viscosity) and
                    is_measurement_good(v.ref_temp)):
                return 0.5

        for v in dvis:
            if (is_measurement_good(v.viscosity) and
                    is_measurement_good(v.ref_temp)):
                return 0.5

    return 0.0


def check_second_viscosity(oil):
    '''
    Fresh oil: Second viscosity at a different temperature.

    Score = maxDeltaT/40, but not greater than 0.5

    maxDeltaT: The difference between the lowest and highest
    measurement in the set.
    '''
    if len(oil.sub_samples) > 0:
        ss = oil.sub_samples[0]
        kvis = ss.physical_properties.kinematic_viscosities
        dvis = ss.physical_properties.dynamic_viscosities

        temps = []
        for v_i in (kvis, dvis):
            temps.extend([v.ref_temp.converted_to('C').value
                          for v in v_i
                          if v.ref_temp is not None])

        if len(temps) >= 2:
            t1, *_, t2 = sorted([t for t in temps if t is not None])
            delta_t = t2 - t1

            if delta_t > 0.0:
                return min(delta_t / 40.0, 0.5)

    return 0.0


def check_distillation(oil):
    '''
    Fresh oil: Two Distillation cuts separated by mass or volume fraction.

    Score = 3 * maxDeltaFraction

    maxDeltaFraction: The difference between the lowest and
    highest measurement in the set
    '''
    if len(oil.sub_samples) > 0:
        ss = oil.sub_samples[0]
        cuts = ss.distillation_data.cuts

        fractions = [c.fraction.converted_to('fraction').value for c in cuts]

        if len(fractions) >= 2:
            f1, *_, f2 = sorted([f for f in fractions if f is not None])

            return 3.0 * (f2 - f1)

    return 0.0


def check_weathered_density(oil):
    '''
        One Evaporated oil: Density. Score = 1
    '''
    if len(oil.sub_samples) > 1:
        ss = oil.sub_samples[1]
        densities = ss.physical_properties.densities

        if (len(densities) > 0 and
                is_measurement_good(densities[0].density) and
                is_measurement_good(densities[0].ref_temp)):
            return 1.0

    return 0.0


def check_weathered_viscosity(oil):
    '''
        One Evaporated oil: Viscosity. Score = 1
    '''
    if len(oil.sub_samples) > 1:
        ss = oil.sub_samples[1]
        kvis = ss.physical_properties.kinematic_viscosities
        dvis = ss.physical_properties.dynamic_viscosities

        if (len(kvis) > 0 and
                is_measurement_good(kvis[0].viscosity) and
                is_measurement_good(kvis[0].ref_temp)):
            return 1.0

        if (len(dvis) > 0 and
                is_measurement_good(dvis[0].viscosity) and
                is_measurement_good(dvis[0].ref_temp)):
            return 1.0

    return 0.0


def is_measurement_good(measurement):
    return not any([(getattr(measurement, a, None) is None)
                    for a in ('value', 'unit', 'unit_type')])


# build a list of all the check function:
CHECKS = [val for name, val in vars().items() if name.startswith("check_")]



"""
cleanups that work with density
"""
import math
from operator import itemgetter

import unit_conversion as uc

from .cleanup import Cleanup

from ....computation.physical_properties import Density

class FixAPI(Cleanup):
    """
    adds (or replaces) the API value, from the density measurements

    NOTE: this could be extended to interpolate, but it that actually needed?
          There is code in the computation.physical_properties package to help, if needed.
    """
    ID = "001"

    def check(self):
        """
        checks to see if there is something to fix

        returns: flag, msg

        if nothing is needed, flag is None
        if something can be cleaned up, flag is True
        if something is wrong, but can not be cleaned up, flag is False

        fixme: -- maybe cleanup and validation should be better integrated?
        """
        API = self.oil.metadata.API

        # densities = oil.sub_samples[0].physical_properties.densities
        if API is None:

            density = self.find_density_near_15C()
            if density:
                return (True, f"Cleanup: {self.ID}: No API value provided for {self.oil.oil_id}"
                               " -- can be computed from density")
            else:
                return (False, f"Cleanup: {self.ID}: No API value provided for {self.oil.oil_id}"
                                " -- can NOT be computed from density")

        return None, "API is fine"

    def cleanup(self):
        """
        run this particular cleanup option

        :param oil: an Oil object to act on

        :param do_it=False: flag to tell the cleanup to do its thing. If False,
                            the method returns a message. If True, the action is
                            taken, and the Oil object is altered.

        :returns: a message of what could be done, or what was done.
        """
        density_at_15 = self.find_density_near_15C()

        if density_at_15:
            API = uc.convert("density", "kg/m^3", "API", density_at_15)
            self.oil.metadata.API = round(API, 2)
            return f"Cleanup: {self.ID}: Set API for {self.oil.oil_id} to {API}."

    def check_for_valid_api(self):
        """
        check is the API value is already valid
        """
        API = self.oil.metadata.API

        density_at_15 = self.find_density_near_15C()

        if uc.convert("density", "kg/m^3", "API", density_at_15) == API:
            return True
        else:
            return False

    def find_density_near_15C(self):
        """
        Returns the density (in kg/m3)

        It will interpolate and extrapolate as needed
        """
        try:
            return Density(self.oil).at_temp(uc.convert("C", "K", 15))
        except ValueError:
            return None



"""
dataclass to store the compounds
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List

from ..common.measurement import MassOrVolumeFraction


@dataclass_to_json
@dataclass
class BulkComposition:
    '''
    object to hold bulk composition -- could be mass or volume fraction
    '''
    name: str = ""
    groups: list = field(default_factory=list)
    method: str = ""
    measurement: MassOrVolumeFraction = None
    comment: str = ""


class BulkCompositionList(JSON_List):
    item_type = BulkComposition



"""
dataclass to store the industry properties

These are tricky, as they can be any weird unit
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List

from ..common.measurement import AnyUnit


@dataclass_to_json
@dataclass
class IndustryProperty:
    '''
    This handles the various odd industry properties.

    For example:

    * Total Acid Number
    * Reid Vapor Pressure
    * Aniline Point
    * Cetane Index
    * Cloud Point
    * Smoke Point
    * Freeze Point
    '''
    name: str = ""
    method: str = ""
    measurement: AnyUnit = None
    comment: str=""

class IndustryPropertyList(JSON_List):
    item_type = IndustryProperty




"""

Classes for individual classes -not much here!

"""

from ..common.utilities import dataclass_to_json
from ..common.validators import YearValidator

from .validation.warnings import WARNINGS
from .validation.errors import ERRORS

from dataclasses import dataclass


@dataclass_to_json
@dataclass
class ReferenceBase:
    year: int = None
    reference: str = ""


class Reference(ReferenceBase):
    _validator = YearValidator(1600,
                               2050,
                               ERRORS["E012"],
                               )

    @classmethod
    def validate(cls, value):
        year = value.year
        if not year:
            return [WARNINGS["W008"]]
        return cls._validator(year)




import sys
import os
import numpy as np
import pandas as pd
import zipfile
import importlib
import traceback
import tkinter as tk
from tkinter import ttk
import adios_db.scripting as ads
from adios_db.models.oil.physical_properties import DynamicViscosityPoint


def parse_weathering_table(df, oil):
    max_water_cont = None
    viscosity_of_max_water_cont = None
    try:
        for i, row in df.iterrows():
            print(row, 'ROW')
            if 'weight residue' in row[0].lower():
                for c in range(1, 5):
                    mo = ads.MassFraction(value=100-float(row[c]), unit="%")
                    oil.sub_samples[c-1].metadata.fraction_evaporated = mo
            elif 'density' in row[0].lower():
                for c in range(1, 5):
                    dp = ads.DensityPoint(
                            density=ads.Density(value=float(row[c]), unit="g/ml"),
                            ref_temp=ads.Temperature(value=13, unit='C'))
                    oil.sub_samples[c-1].physical_properties.densities.append(dp)
            elif 'viscosity' in row[0].lower() and 'free' in row[0].lower():
                for c in range(1, 5):
                    dp = DynamicViscosityPoint(
                            viscosity=ads.DynamicViscosity(value=float(row[c]), unit="cP"),
                            ref_temp=ads.Temperature(value=13, unit='C'))
                    oil.sub_samples[c-1].physical_properties.dynamic_viscosities.append(dp)
            elif 'viscosity' in row[0].lower() and 'max' in row[0].lower():
                viscosity_of_max_water_cont = row[1:5]
            elif 'max' in row[0].lower() and 'water cont' in row[0].lower():
                max_water_cont = row[1:5]
    except:
        raise ValueError('Not a weathering table')

    if viscosity_of_max_water_cont is not None and max_water_cont is not None:
        for c, visc, watcont in zip(range(1, 5), viscosity_of_max_water_cont, max_water_cont):
            try:
                eo = ads.Emulsion(water_content=ads.MassFraction(value=float(watcont), unit='%'),
                                  ref_temp=ads.Temperature(value=13, unit='C'),
                                  complex_viscosity=ads.DynamicViscosity(value=float(visc), unit='cP'))
                oil.sub_samples[c-1].environmental_behavior.emulsions.append(eo)
                print(c, eo)
            except ValueError:
                pass
    else:
        logfile.write('Discarding whole table\n')
        return

    print(oil, 'Finished oil!')

class ParseOilPDF(tk.Tk):

    def __init__(self):
        f = sys.argv[1]  # input file name
        b = f.split('.')[0]

        if not os.path.exists(b+'.zip'):
            print('Parsing %s, this may take some minutes...' % f)
            import camelot.io as camelot
            tables = camelot.read_pdf(f, pages='all')
            print('Total tables extracted: %s\n' % tables.n)
            tables.export(b+'.csv', f='csv', compress=True)

        z = zipfile.ZipFile(b+'.zip')
        print('Number of tables: %s\n' % len(z.namelist()))

        tk.Tk.__init__(self)
        self.title('Oil parser for Sintef PDF reports')

        # Make tabs
        self.n = ttk.Notebook(self.master)
        self.n.grid()
        self.tabs = ttk.Frame(self.n)
        self.oil = ttk.Frame(self.n)
        self.n.add(self.tabs, text='Tables')
        self.n.add(self.oil, text='Oil')

        mf = tk.Frame(self, bg='cyan', width=1200, height=150, padx=3, pady=3)
        cf = tk.Frame(self, bg='gray2', width=1200, height=900, padx=3, pady=3)

        # layout all of the main containers
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        mf.grid(row=0, sticky="nsew")
        cf.grid(row=1, sticky="nsew")

        # create the center widgets
        cf.grid_rowconfigure(0, weight=1)
        cf.grid_columnconfigure(1, weight=1) 

        height = 600
        properties = tk.Frame(cf, bg='gray', padx=3, pady=3)
        cuts = tk.Frame(cf, bg='red', padx=3, pady=3)
        weathering = tk.Frame(cf, bg='blue', padx=3, pady=3)

        properties.grid(row=0, column=0, sticky='nsew')
        cuts.grid(row=0, column=1, sticky='nsew')
        weathering.grid(row=0, column=2, columnspan=4, sticky='nsew')

        self.tables = {}
        self.dataframes = {}
        self.selected_tables = []
        tableno = 0
        for table in z.namelist():
            df = pd.read_csv(z.open(table))
            if df.iloc[:, 1:].isnull().all().all():
                print('Skipping empty table\n')
                continue  # Do not print empty tables
            if len(df.columns) not in [2, 5]:
                print('Skipping table without 2 or 5 columns\n')
                continue

            color='lightgray'
            if len(df.columns) == 2:
                if 'temp' in df.columns[0].lower() and (
                        'vol' in df.columns[0].lower() or 'vol' in df.columns[1].lower()):
                    parent = cuts
                    width=60
                else:
                    parent = properties
                    width=60
            elif len(df.columns) == 5:
                # Check if this is a valid weathering table
                oil = ads.Oil('dummy')
                oil.sub_samples.extend([ads.Sample(), ads.Sample(), ads.Sample(), ads.Sample()])
                for i, s in enumerate(['Fresh oil', 'Topped to 150C', 'Topped to 200C', 'Topped to 250C']):
                    oil.sub_samples[i].metadata.name = s
                    oil.sub_samples[i].metadata.short_name = s
                try:
                    parse_weathering_table(df, oil)
                except Exception as e:
                    print(e)
                    print(traceback.format_exc())
                    print('Not a weathering table')
                    continue

                parent = weathering
                width=150

            self.dataframes[table] = df
            self.tables[table] = tk.Text(parent, bg=color, padx=3, pady=3, width=width)
            tableno = tableno + 1
            self.tables[table].insert(tk.END, df)
            self.tables[table].grid(row=tableno, column=0, sticky='ns')
            self.tables[table].bind('<Button-1>', lambda event,s=table:self.clicktable(s))

    def clicktable(self, table):
        if table in self.selected_tables:  # Remove/unselect
            self.tables[table].config(bg='lightgray')
            self.selected_tables.remove(table)
        else:
            self.tables[table].config(bg='green')
            self.selected_tables.append(table)

        print(self.selected_tables)
        print('Making oil')
        oil = ads.Oil('dummy')
        oil.sub_samples.extend([ads.Sample(), ads.Sample(), ads.Sample(), ads.Sample()])
        for i, s in enumerate(['Fresh oil', 'Topped to 150C', 'Topped to 200C', 'Topped to 250C']):
            oil.sub_samples[i].metadata.name = s
            oil.sub_samples[i].metadata.short_name = s
        print(oil)
        for tablename in self.selected_tables:
            df = self.dataframes[tablename]
            print('-'*22)
            print(df, 'DF')
            print('-'*22)
            try:
                parse_weathering_table(df, oil)
            except Exception as e:
                print(e)
                print(traceback.format_exc())

        oil.to_file('test.json')
        v = oil.validate()
        print(v, 'VALIDATE')
        import os
        os.system('cat test.json')

if __name__ == '__main__':
    ParseOilPDF().mainloop()



########################################################
## Step 1: the pdf is parsed and cached in a .zip file
########################################################
#if suffix == 'pdf':
#    if not os.path.exists(b+'.zip'):
#        logfile.write('Parsing ' + f)
#        import camelot.io as camelot
#        tables = camelot.read_pdf(f, pages='all')
#        logfile.write('Total tables extracted: %s\n' % tables.n)
#        tables.export(b+'.csv', f='csv', compress=True)
#
#    z = zipfile.ZipFile(b+'.zip')
#
#    for table in z.namelist():
#        df = pd.read_csv(z.open(table))
#        logfile.write('*'*33 + '\n')
#        logfile.write(str(df) + '\n')
#        logfile.write('-'*33 + '\n')
#        if df.iloc[:, 1:].isnull().all().all():
#            logfile.write('Skipping empty table\n')
#            continue  # Do not print empty tables
#        if len(df.columns) not in [2, 5]:
#            logfile.write('Skipping table without 2 or 5 columns\n')
#            continue
#        df = df.astype(str)
#        df = df.replace(r'\n',' ', regex=True)
#        headers = df.columns
#        if '\n' in headers[0]:
#            headers = headers[0].split('\n')
#            if len(headers) > len(df.columns):
#                headers = headers[0:len(df.columns)]
#            elif len(headers) < len(df.columns):
#                headers = df.columns
#                headers = [h.replace('\n', '') for h in headers]
#        maxcollengths = [df[h].astype(str).str.len().max() for h in df.columns]
#        formats = ['{:>%ss}%s' % (m,sep) for m in maxcollengths]
#        formats[0] = formats[0].replace('>', '<')
#        formats[-1] = formats[-1].replace(sep, '')
#        formats = [f.format for f in formats]
#        hs = ''
#        for h,f in zip(headers, formats):
#            hs = hs + f(h.replace('\n', ''))
#        logfile.write(str(hs) + '\n')
#        outfile.write(df.to_string(index=False, header=True, formatters=formats) + '\n\n')
#        logfile.write('*'*33 + '\n')
#
#    metadata = {
#        'Key': 'Value',
#        'oilID': 'oilID',
#        'title': 'title',
#        'authors': 'authors',
#        'region': 'region',
#        }
#    
#    #print('# Metadata')
#    for i, (key, value) in enumerate(metadata.items()):
#        outfile.write('%s%s%s\n' % (key, sep, value))
#    outfile.write('\n')
#
###########################################
## Step 3: parsing tables file to create
##         oil object, and export to json
###########################################
#elif suffix == 'tables':
#
#    logfile.write('='*33 + '\n')
#    oil = ads.Oil('dummy')
#    oil.sub_samples.extend([ads.Sample(), ads.Sample(), ads.Sample(), ads.Sample()])
#    for i, s in enumerate(['Fresh oil', 'Topped to 150C', 'Topped to 200C', 'Topped to 250C']):
#        oil.sub_samples[i].metadata.name = s
#        oil.sub_samples[i].metadata.short_name = s
#
#    #from pandas.compat import StringIO #if this doesn't work try: from io import StringIO
#    from io import StringIO
#    logfile.write('Parsing tables file\n')
#    tables = open(f).read().split('\n\n')
#    logfile.write('Number of tables: %s\n' % len(tables))
#    for i, t in enumerate(tables):
#        #print(t)
#        #print(i)
#        #title = t.split('\n')[0][2:]
#        #t = '\n'.join(t.split('\n')[1:])
#        df = pd.read_csv(StringIO(t), sep=sep)
#        logfile.write(df, '\n\n')
#        stapp
#
#        # Parse table here, into oil object?
#        try:
#            parse_weathering_table(df, oil)
#        except:
#            logfile.write('Not a weathering table\n')
#
#    #print(tables, 'TAB')
#    logfile.write('Finished_parsing_tables\n')
#    jolla
#
#results = {}  # to store the parsed values
#z = zipfile.ZipFile(b+'.zip')
#logfile.write('Parsing zipfile ' + b+'.zip\n')
#for table in z.namelist():
#    #wd = {}
#    #print('Parsing table ' + table)
#    df = pd.read_csv(z.open(table))
#    if df.isnull().all().all():
#        continue  # empty table
#    logfile.write('# %s\n' % table)
#    logfile.write('%s\n\n' % df.to_string())
#    continue
#    if table != 'TOR_oil-page-95-table-1.csv':
#        continue
#
#    parse_weathering_table(df, oil)
#    parsing
#
#    # Search for row parameters
#    for v, r in template['rows'].items():
#        for name in r['names']:
#            for i, row in df.iterrows():
#                if name in str(row):
#                    varname = v
#                    while(varname) in results:
#                        varname = varname + 'a'
#                    results[varname] = []
#                    #print(varname + str(row.values))
#                    for ri in range(len(row.values)):
#                        try:
#                            row.values[ri] = np.float32(row.values[ri].replace(',', '.'))
#                        except:
#                            pass
#                    try:
#                        for aval in row.values:
#                            try:
#                                results[varname].append(np.float32(aval))
#                            except:
#                                pass
#                        results[varname] = np.array(results[varname])
#                        if 'factor' in r:
#                            results[varname] = results[varname]*r['factor']
#                        if 'subtract_from' in r:
#                            results[varname] = r['subtract_from'] - results[varname]
#                        if 'add_to' in r:
#                            results[varname] = r['add_to'] + results[varname]
#                        #print(results[varname])
#                    except:
#                        logfile.write('Could not parse row: %s\n' % str(row))
#                        stop
#                        pass
#    stop
#
#    #print('='*30)
#    #print(df)
#    #print('='*30)
#    for v, r in template['columns'].items():
#        headerline = df.columns
#        if 'Temp' in headerline[0] and ('Vol' in headerline[1] or 'Duva' in headerline[1]):
#            cut_temps = df.iloc[:,0].values + 273
#            cut_vols = df.iloc[:,1].values
#            try:
#                cut_vols = cut_vols / 100
#            except:
#                cut_vols = [np.float64(val.replace(',', '.'))/100 for val in cut_vols]
#
#            re = np.empty(len(cut_temps)*3)
#            for n in range(len(cut_temps)):
#                re[n*3] = cut_temps[n]
#                re[n*3+2] = cut_vols[n]
#            logfile.write('Cuts:\n' + np.array2string(re.astype(np.float32), separator='\t').replace('\n', '\t') + '\n')
##print('='*30)
##print(pd.read_csv(z.open(z.namelist()[0])))
##print('='*30)
#for n, r in results.items():
#    if len(r) > 0 and np.sum(np.isfinite(r))>0:
#        logfile.write(n, r, '\n')
#logfile.write('='*30 + '\n')
#
#for aas in ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']:
#    dstring = 'density' + aas
#    vstring = 'viscosity' + aas
#    wstring = 'weathered' + aas
#    rd = np.empty(12)
#    rv = np.empty(12)
#    for i in range(4):
#        if dstring in results and len(results[dstring])>0:
#            rd[i*3] = results[dstring][i]
#            rd[i*3+1] = 286
#            rd[i*3+2] = results[wstring][i]
#        if vstring in results and len(results[vstring])>0:
#            rv[i*3] = results[vstring][i]
#            rv[i*3+1] = 286
#            rv[i*3+2] = results[wstring][i]
#
#    if vstring in results:
#        logfile.write(vstring, np.array2string(rv.astype(np.float32), separator='\t').replace('\n', '\t')+ '\n')
#    if dstring in results:
#        logfile.write(dstring, np.array2string(rd.astype(np.float32), separator='\t').replace('\n', '\t') + '\n')
#
#    
##try:
##    r = np.empty(12)
##    for i in range(4):
##        r[i*3] = results['density'][i]
##        r[i*3+1] = 286
##        r[i*3+2] = results['weathered'][i]
##    print(np.array2string(r.astype(np.float32), separator='\t').replace('\n', '\t'))
##except:
##    print('No density')
##
##try:
##    for i in range(4):
##        r[i*3] = results['viscosity'][i]
##    print(np.array2string(r.astype(np.float32), separator='\t').replace('\n', '\t'))
##except:
##    print('No viscosity')



#!/usr/bin/env python
#
# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway


# Utility script to display contents of a netCDF CF-compliant
# file or URL containing driver data suitable for opendrift
#
# Knut-Frode Dagestad, 19 Feb 2015

import sys
import argparse
from datetime import datetime

import numpy as np
import matplotlib.pyplot as plt

sys.path.append("..")

try:
    from opendrift.readers import reader_netCDF_CF_generic
    from opendrift.readers import reader_ROMS_native
    readers = [reader_netCDF_CF_generic, reader_ROMS_native]
except ImportError: # development
    sys.exit('Please add opendrift folder to your PYTHONPATH.')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename',
                        help='<URL or netCDF filename>')
    parser.add_argument('lon', help='longitude in degrees')
    parser.add_argument('lat', help='longitude in degrees')
    parser.add_argument('time', help='YYYYMMDDHHMM')
    parser.add_argument('-e', action='store_true',
                        help='Report errors on failure.')

    args = parser.parse_args()

    for reader in readers:
        try:
            print('Testing %s...' % reader.__file__)
            r = reader.Reader(args.filename)
            print(r)
            break
        except Exception as me:
            if args.e is True:
                print(me)
                import traceback
                print(traceback.format_exc())
                print('---------------------------------------')
            print('...not applicable.')

    if not 'r' in locals():
        sys.exit('No readers applicable for ' + args.filename)

    time = datetime.strptime(args.time, '%Y%m%d%H%M')
    lon = np.atleast_1d(args.lon)
    lat = np.atleast_1d(args.lat)

    x, y = r.lonlat2xy(lon, lat)
    r.buffer = 3
    i=3; j=3  # center of block

    uv = r.get_variables(['x_sea_water_velocity', 'y_sea_water_velocity',
                         'sea_floor_depth_below_sea_level'],
                         time, x, y, r.z)

    u_comp = uv['x_sea_water_velocity'][:,i,j]
    v_comp = uv['y_sea_water_velocity'][:,i,j]
    depth = uv['sea_floor_depth_below_sea_level'][i,j]
    u_comp = u_comp[0:sum(~u_comp.mask)]
    v_comp = v_comp[0:sum(~v_comp.mask)]

    u_rot, v_rot = r.rotate_vectors(x, y, u_comp, v_comp,
                                    r.proj, '+proj=latlong')

    vel = np.sqrt(u_rot**2 + v_rot**2)
    mx = np.nanmax((np.abs(u_rot), np.abs(v_rot)))
    fig = plt.figure()
    ax = fig.gca()
    plt.quiver(np.zeros(len(u_rot)), np.zeros(len(u_rot)), u_rot, v_rot, r.z[0:sum(~u_rot.mask)],
               scale=1.1, scale_units='x')
    plt.axis([-mx, mx, -mx, mx])
    plt.grid('on')
    cb = plt.colorbar(ticks=r.z[0:sum(~u_rot.mask)])
    cb.set_label('Depth [m]')
    plt.ylabel('North component [m/s]')
    plt.xlabel('East component [m/s]')
    textstr = u'%s UTC\n%.3fN\N{DEGREE SIGN}, %.3fE\N{DEGREE SIGN}' % \
             (time, float(args.lat), float(args.lon))
    ax.text(0.05, 0.95, textstr, transform=ax.transAxes, fontsize=14,
            verticalalignment='top',
            bbox=dict(boxstyle='round', facecolor='wheat', alpha=0.5))
    plt.show()

if __name__ == '__main__':
    main()




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from collections import OrderedDict
import numpy as np

class LagrangianArray:
    """A generic array-like class for Lagrangian particle tracking.

    A LagrangianArray is a generic class keeping the values of given
    properties ('variables') of a collection of particles at a given time.
    Values are stored as named attributes (similar to recarray) which are
    ndarrays (1D, vectors) with one value for each particle, or as scalars
    for values shared among all particles.

    This is an Abstract Base Class, meaning that only subclasses can be used.
    Subclasses will add specific variables for specific purposes (particle
    types, e.g. oil, fish eggs...) to the core variables described below.

    Attributes:

        variables: An OrderedDict where keys are names of the
            variables/properties of the current object. The values
            of the OrderedDict are dictionaries with names such as
            'dtype', 'unit', 'standard_name' (CF), 'default' etc.
            All variable names will be added dynamically as attributes of
            the object after initialisation. These attributes will be
            numpy ndarrays of same length, or scalars. The core variables
            are:

                - ID: an integer identifying each particle.
                - status: 0 for active particles and a positive integer when deactivated
                - lon: longitude (np.float32)
                - lat: latitude (np.float32)
                - z: vertical position of the particle in m, positive upwards (above sea surface)
    """

    variables = OrderedDict([
        ('ID', {'dtype': np.int32,  # Unique numerical identifier
                'seed': False,
                'default': -1}),  # ID to be assigned by application
        ('status', {'dtype': np.int32,  # Status categories
                    'seed': False,
                    'default': 0}),
        ('moving', {'dtype': np.int32,  # Set to 0 for elements which are frosen
                    'seed': False,
                    'default': 1}),
        ('age_seconds', {'dtype': np.float32,
                         'units': 's',
                         'seed': False,
                         'default': 0}),
        ('origin_marker', {'dtype': np.int16,
                           'unit': '',
            'description': 'An integer kept constant during the simulation. Different values may be used for different seedings, to separate elements during analysis. With GUI, only a single seeding is possible.',
                           'default': 0}),
        ('lon', {'dtype': np.float32,
                 'units': 'degrees_east',
                 'standard_name': 'longitude',
                 'long_name': 'longitude',
                 'seed': False,
                 'axis': 'X'}),
        ('lat', {'dtype': np.float32,
                 'units': 'degrees_north',
                 'standard_name': 'latitude',
                 'long_name': 'latitude',
                 'seed': False,
                 'axis': 'Y'}),
        ('z', {'dtype': np.float32,
                   'units': 'm',
                   'standard_name': 'z',
                   'long_name': 'vertical position',
                   'axis': 'Z',
                   'positive': 'up',
                   'default': 0})])

    def __init__(self, **kwargs):
        """Initialises a LagrangianArray with given properties.

        Args:
            Keyword arguments (kwargs) with names corresponding to the
            OrderedDict 'variables' of the class, and corresponding values.
            The values must be ndarrays of equal length, or scalars.
            All (or none) variables must be given, unless a default value
            is specified in the OrderedDict 'variables'
            An empty object may be created by giving no input.
        """

        # Collect default values in separate dict, for easier access
        default_values = {variable: self.variables[variable]['dtype'](
                          self.variables[variable]['default'])
                          for variable in self.variables
                          if 'default' in self.variables[variable]}

        if len(kwargs) == 0:
            # Initialise an empty object (all variables have length 0)
            for var in self.variables:
                kwargs[var] = []

        # Check for missing arguments
        missing_args = set(self.variables.keys()) - \
            set(kwargs.keys()) - set(default_values.keys())
        if missing_args:
            raise TypeError('Missing arguments: %s' % str(list(missing_args)))

        # Check for redundant arguments
        redundant_args = set(list(kwargs.keys()) +
            list(default_values.keys())) - set((self.variables.keys()))
        if redundant_args:
            raise TypeError('Redundant arguments: %s' %
                            str(list(redundant_args)))

        # Check that input arrays have same length
        array_lengths = [1]*len(kwargs)
        for i, input_value in enumerate(kwargs.values()):
            try:
                array_lengths[i] = len(input_value)
            except:
                array_lengths[i] = 1  # scalar is given
        if len(set(array_lengths) - {1}) > 1:
            raise TypeError(
                'Input arrays must have same length. Lengths given: '
                + str(array_lengths))

        # Store input arrays
        for default_variable in default_values.keys():  # set default values
            setattr(self, default_variable, default_values[default_variable])
        for input_variable in kwargs.keys():  # override with input values
            setattr(self, input_variable, self.variables[input_variable]
                    ['dtype'](kwargs[input_variable]))

        # Store dtypes for all parameters in a common dtype object (for io)
        self.dtype = np.dtype([(var[0], var[1]['dtype'])
                               for var in self.variables.items()])

        # Status must always be array
        if not type(self.status) == np.ndarray:
            self.status = self.status*np.ones(self.lon.shape)

    @classmethod
    def add_variables(cls, new_variables):
        """Method used by subclasses to add specific properties/variables."""
        variables = cls.variables.copy()
        variables.update(new_variables)
        return variables

    def extend(self, other):
        """Add elements from another object."""
        len_self = len(self)
        len_other = len(other)
        for var in self.variables:
            present_data = getattr(self, var)
            new_data = getattr(other, var)

            # If both arrays have an identical scalar, it remains a scalar
            if (not isinstance(new_data, np.ndarray) and
                not isinstance(present_data, np.ndarray) and
                present_data == new_data):
                continue

            else:  # Otherwise we create arrays and concatenate
                if not hasattr(present_data, '__len__'):
                    present_data = present_data*np.ones(len_self)
                if not hasattr(new_data, '__len__'):
                    new_data = new_data*np.ones(len_other)
                setattr(self, var, np.concatenate((present_data,
                                                   new_data)))

    def move_elements(self, other, indices):
        """Remove elements with given indices, and append to another object.
        NB: indices is boolean array, not real indices!"""

        # Move elements with given indices (boolean array)
        # to another LagrangianArray
        # NB: scalars and 1D arrays are converted to ndarrays and concatenated
        self_len = len(self)
        other_len = len(other)
        for var in self.variables:
            self_var = getattr(self, var)
            other_var = getattr(other, var)
            if (not isinstance(self_var, np.ndarray) and
                not isinstance(other_var, np.ndarray)) and \
                    (other_var == self_var):
                    if np.sum(indices) == len(self):
                        setattr(self, var, [])  # Empty if all elements moved
                    continue  # Equal scalars - we do nothing

            # Copy elements to other
            self_var = np.atleast_1d(self_var)
            other_var = np.atleast_1d(other_var)
            if len(self_var) < self_len:  # Convert scalar to array
                self_var = self_var*np.ones(self_len)
            if len(other_var) < other_len:  # Convert scalar to aray
                other_var = other_var*np.ones(other_len)
            if len(self_var) > 0:
                setattr(other, var, np.concatenate((other_var,
                                                    self_var[indices])))
            else:
                setattr(other, var, self_var[indices])
            setattr(self, var, self_var[~indices])  # Remove from self

            #if isinstance(self_var, np.ndarray) or\
            #    isinstance(other_var, np.ndarray):  # Array
            #    setattr(other, var,
            #            np.concatenate((getattr(other, var),
            #                            getattr(self, var)[indices])))
            #    # Remove elements from self
            #    setattr(self, var, getattr(self, var)[~indices])
            #elif isinstance(getattr(other, var), np.ndarray):
            #    setattr(other, var,
            #            np.concatenate((getattr(other, var),
            #                            np.atleast_1d(getattr(self, var)))))
            #else:
            #    setattr(other, var, getattr(self, var))  # Scalar

    def __len__(self):
        length = 0
        for var in self.variables:
            length = np.maximum(length, len(np.atleast_1d(getattr(self, var))))
        return length

    def __repr__(self):
        outStr = ''
        for variable in self.variables.keys():
            outStr += variable + ': ' + str(getattr(self, variable)) + '\n'
        return outStr



import numpy as np
import pyproj
from scipy.ndimage import map_coordinates
from abc import abstractmethod

from opendrift.readers.interpolation.structured import ReaderBlock
from .variables import Variables

import logging
logger = logging.getLogger(__name__)


class StructuredReader(Variables):
    """
    A structured reader. Data is gridded on a regular grid. Used by e.g.:
    :class:`opendrift.readers.reader_netCDF_CF_generic.Reader`.

    Attributes:

        projected: is `True` if :class:`.fakeproj.fakeproj` is used because of missing projection information. The data points are assumed to be approximately equidistant on the surface (i.e. in meters).

        clipped: pixels to to remove along boundary (e.g. in case of bad data).

    .. seealso::

        :py:mod:`opendrift.readers`
    """

    # TODO: should the variables below not be instance variables, and not class variables?
    clipped = 0
    x = None
    y = None
    interpolation = 'linearNDFast'
    convolve = None  # Convolution kernel or kernel size

    # Used to enable and track status of parallel coordinate transformations.
    __lonlat2xy_parallel__ = None
    __disable_parallel__ = False

    def __init__(self):
        if self.proj is None and (self.proj4 is None
                                  or self.proj4 == 'fakeproj'):

            logger.warning(
                "No proj string or projection could be derived, using 'fakeproj'. This assumes that the variables are structured and gridded approximately equidistantly on the surface (i.e. in meters). This must be guaranteed by the user. You can get rid of this warning by supplying a valid projection to the reader."
            )

            from scipy.interpolate import LinearNDInterpolator
            import copy
            from . import fakeproj

            # `projected` is set to True if `fakeproj` is used
            self.projected = None
            self.shape = None

            self.proj4 = 'None'
            self.proj = fakeproj.fakeproj()
            self.projected = False
            logger.info('Making interpolator for lon,lat to x,y conversion...')
            self.xmin = self.ymin = 0.
            self.delta_x = self.delta_y = 1.
            self.xmax = self.lon.shape[1] - 1
            self.ymax = self.lon.shape[0] - 1
            self.numx = self.xmax
            self.numy = self.ymax
            self.x = np.arange(0, self.xmax+1)
            self.y = np.arange(0, self.ymax+1)

            block_x, block_y = np.mgrid[self.xmin:self.xmax + 1,
                                        self.ymin:self.ymax + 1]
            block_x, block_y = block_x.T, block_y.T

            # Making interpolator (lon, lat) -> x
            self.spl_x = LinearNDInterpolator(
                (self.lon.ravel(), self.lat.ravel()),
                block_x.ravel(),
                fill_value=np.nan)
            # Reusing x-interpolator (deepcopy) with data for y
            self.spl_y = copy.deepcopy(self.spl_x)
            self.spl_y.values[:, 0] = block_y.ravel()
            # Call interpolator to avoid threading-problem:
            # https://github.com/scipy/scipy/issues/8856
            self.spl_x((0, 0)), self.spl_y((0, 0))
        else:
            self.projected = True

        super().__init__()

        # Dictionaries to store blocks of data for reuse (buffering)
        self.var_block_before = {}  # Data for last timestep before present
        self.var_block_after = {}  # Data for first timestep after present

    @abstractmethod
    def get_variables(self, variables, time=None, x=None, y=None, z=None):
        """
        Obtain a _block_ of values of the requested variables at all positions
        (x, y, z) closest to given time. These will be stored in
        :class:`opendrift.readers.interpolation.structured.ReaderBlock` and
        accessed from there.

        Arguments:
            variables: list of variables.

            time: datetime or None, time at which data are requested.

            x, y: float or ndarrays; coordinates of requested points.

            z: float or ndarray; vertical position (in meters, positive up)

          Returns:
            Dictionary

            keywords: variables (string)
            values: 2D ndarray bounding x and y.
        """

    def prepare(self, extent, start_time, end_time, max_speed):
        """Prepare reader for given simulation coverage in time and space."""
        logger.debug('Clearing cache for reader %s before starting new simulation' % self.name)
        self.var_block_before = {}
        self.var_block_after = {}
        if self.time_step is None:  # Set buffer large nough for whole simulation
                logger.debug('Time step is None for %s, setting buffer size large nough for whole simulation' % self.name)
                self.set_buffer_size(max_speed, end_time-start_time)
        super().prepare(extent, start_time, end_time, max_speed)

    def set_convolution_kernel(self, convolve):
        """Set a convolution kernel or kernel size (of array of ones) used by `get_variables` on read variables."""
        self.convolve = convolve

    def __convolve_block__(self, env):
        """
        Convolve arrays with a kernel, if reader.convolve is set
        """
        if self.convolve is not None:
            from scipy import ndimage
            N = self.convolve
            if isinstance(N, (int, np.integer)):
                kernel = np.ones((N, N))
                kernel = kernel / kernel.sum()
            else:
                kernel = N
            logger.debug('Convolving variables with kernel: %s' % kernel)
            for variable in env:
                if variable in ['x', 'y', 'z', 'time']:
                    pass
                else:
                    if env[variable].ndim == 2:
                        env[variable] = ndimage.convolve(env[variable],
                                                         kernel,
                                                         mode='nearest')
                    elif env[variable].ndim == 3:
                        env[variable] = ndimage.convolve(env[variable],
                                                         kernel[:, :, None],
                                                         mode='nearest')
        return env

    def lon_range(self):
        if not self.global_coverage():
            raise ValueError('Only valid for readers with global coverage')
        if self.xmin < 0:
            return '-180to180'
        else:
            return '0to360'

    def _get_variables_interpolated_(self, variables, profiles, profiles_depth,
                                     time, reader_x, reader_y, z):

        # For global readers, we shift coordinates to match actual lon range
        if self.global_coverage():
            if self.lon_range() == '-180to180':
                logger.debug('Shifting coordinates to -180-180')
                reader_x = np.mod(reader_x + 180, 360) - 180
            elif self.lon_range() == '0to360':
                logger.debug('Shifting coordinates to 0-360')
                reader_x = np.mod(reader_x, 360)
        elif self.proj.crs.is_geographic and self.xmin>0:
            logger.debug('Modulating longitudes to 0-360 for self.name')
            reader_x = np.mod(reader_x, 360)

        # Find reader time_before/time_after
        time_nearest, time_before, time_after, i1, i2, i3 = \
            self.nearest_time(time)
        logger.debug('Reader time:\n\t\t%s (before)\n\t\t%s (after)' %
                     (time_before, time_after))

        # For variables which are not time dependent, we do not care about time
        static_variables = [
            'sea_floor_depth_below_sea_level', 'land_binary_mask'
        ]
        if time == time_before or all(v in static_variables
                                      for v in variables):
            time_after = None

        if profiles is not None:
            # If profiles are requested for any parameters, we
            # add two fake points at the end of array to make sure that the
            # requested block has the depth range required for profiles
            mx = np.append(reader_x, [reader_x[-1], reader_x[-1]])
            my = np.append(reader_y, [reader_y[-1], reader_y[-1]])
            mz = np.append(z, [profiles_depth[0], profiles_depth[1]])
        else:
            mx = reader_x
            my = reader_y
            mz = z

        block_before = block_after = None
        blockvariables_before = variables
        blockvars_before = str(variables)
        blockvariables_after = variables
        blockvars_after = str(variables)
        for blockvars in self.var_block_before:
            if all(v in blockvars for v in variables):
                block_before = self.var_block_before[blockvars]
                blockvariables_before = block_before.data_dict.keys()
                blockvars_before = blockvars
                break
            blockvariables_before = variables
            blockvars_before = str(variables)
        for blockvars in self.var_block_after:
            if all(v in blockvars for v in variables):
                block_after = self.var_block_after[blockvars]
                blockvariables_after = block_after.data_dict.keys()
                blockvars_after = blockvars
                break

        # Swap before- and after-blocks if matching times
        if block_before is not None and block_after is not None:
            if block_before.time != time_before:
                if block_after.time == time_before:
                    block_before = block_after
                    self.var_block_before[blockvars_before] = block_before
            if block_after.time != time_after:
                if block_before.time == time_before:
                    block_after = block_before
                    self.var_block_after[blockvars_after] = block_after

        # Fetch data, if no buffer is available
        if block_before is None or \
                block_before.time != time_before:
            reader_data_dict = \
                    self.__convolve_block__(
                self.get_variables(blockvariables_before, time_before,
                                    mx, my, mz)
                )
            self.var_block_before[blockvars_before] = \
                ReaderBlock(reader_data_dict,
                            interpolation_horizontal=self.interpolation)
            try:
                len_z = len(self.var_block_before[blockvars_before].z)
            except:
                len_z = 1
            logger.debug(
                ('Fetched env-block (size %ix%ix%i) ' + 'for time before (%s)')
                % (len(self.var_block_before[blockvars_before].x),
                   len(self.var_block_before[blockvars_before].y), len_z,
                   time_before))
            block_before = self.var_block_before[blockvars_before]
        if block_after is None or block_after.time != time_after:
            if time_after is None:
                self.var_block_after[blockvars_after] = block_before
            else:
                reader_data_dict = self.__convolve_block__(
                    self.get_variables(blockvariables_after, time_after, mx,
                                       my, mz))
                self.var_block_after[blockvars_after] = \
                    ReaderBlock(
                        reader_data_dict,
                        interpolation_horizontal=self.interpolation)
                try:
                    len_z = len(self.var_block_after[blockvars_after].z)
                except:
                    len_z = 1

                logger.debug(('Fetched env-block (size %ix%ix%i) ' +
                              'for time after (%s)') %
                             (len(self.var_block_after[blockvars_after].x),
                              len(self.var_block_after[blockvars_after].y),
                              len_z, time_after))
                block_after = self.var_block_after[blockvars_after]

        if (block_before is not None and block_before.covers_positions(
            reader_x, reader_y) is False) or (\
            block_after is not None and block_after.covers_positions(
                reader_x, reader_y) is False):
            logger.warning('Data block from %s not large enough to '
                           'cover element positions within timestep. '
                           'Buffer size (%s) must be increased. See `Variables.set_buffer_size`.' %
                           (self.name, str(self.buffer)))
            # TODO; could add dynamic incraes of buffer size here

        ############################################################
        # Interpolate before/after blocks onto particles in space
        ############################################################
        self.timer_start('interpolation')
        logger.debug('Interpolating before (%s) in space  (%s)' %
                     (block_before.time, self.interpolation))
        env_before, env_profiles_before = block_before.interpolate(
            reader_x, reader_y, z, variables, profiles, profiles_depth)

        if (time_after is not None) and (time_before != time):
            logger.debug('Interpolating after (%s) in space  (%s)' %
                         (block_after.time, self.interpolation))
            env_after, env_profiles_after = block_after.interpolate(
                reader_x, reader_y, z, variables, profiles, profiles_depth)

        self.timer_end('interpolation')

        #######################
        # Time interpolation
        #######################
        self.timer_start('interpolation_time')
        env_profiles = None
        if (time_after is not None) and (time_before != time) and self.always_valid is False:
            weight_after = ((time - time_before).total_seconds() /
                            (time_after - time_before).total_seconds())
            logger.debug(('Interpolating before (%s, weight %.2f) and'
                          '\n\t\t      after (%s, weight %.2f) in time') %
                         (block_before.time, 1 - weight_after,
                          block_after.time, weight_after))
            env = {}
            for var in variables:
                # Weighting together, and masking invalid entries
                env[var] = np.ma.masked_invalid(
                    (env_before[var] * (1 - weight_after) +
                     env_after[var] * weight_after))
            # Interpolating vertical profiles in time
            if profiles is not None:
                env_profiles = {}
                logger.debug('Interpolating profiles in time')
                # Truncating layers not present both before and after
                numlayers = np.minimum(len(env_profiles_before['z']),
                                       len(env_profiles_after['z']))
                env_profiles['z'] = env_profiles_before['z'][0:numlayers]
                for var in env_profiles_before.keys():
                    if var == 'z':
                        continue
                    env_profiles_before[var] = np.atleast_2d(
                        env_profiles_before[var])
                    env_profiles_after[var] = np.atleast_2d(
                        env_profiles_after[var])
                    env_profiles[var] = (
                        env_profiles_before[var][0:numlayers, :] *
                        (1 - weight_after) +
                        env_profiles_after[var][0:numlayers, :] * weight_after)
            else:
                env_profiles = None

        else:
            logger.debug('No time interpolation needed - right on time.')
            env = env_before
            if profiles is not None:
                if 'env_profiles_before' in locals():
                    env_profiles = env_profiles_before
                else:
                    # Copying data from environment to vertical profiles
                    env_profiles = {'z': profiles_depth}
                    for var in profiles:
                        env_profiles[var] = np.ma.array([env[var], env[var]])
        self.timer_end('interpolation_time')

        return env, env_profiles

    def __check_env_arrays__(self, env):
        """
        For the StructuredReader the variables are checked before entered into
        the ReaderBlock interpolator. This methods makes the second check a
        no-op.

        .. seealso::

            :meth:`.variables.Variables.__check_env_arrays__`.
        """
        return env

    def xy2lonlat(self, x, y):
        if self.projected:
            return super().xy2lonlat(x, y)
        else:
            np.seterr(invalid='ignore')  # Disable warnings for nan-values
            y = np.atleast_1d(y)
            x = np.atleast_1d(x)

            # NB: mask coordinates outside domain
            x[x < self.xmin] = np.nan
            x[x > self.xmax] = np.nan
            y[y < self.ymin] = np.nan
            y[y < self.ymin] = np.nan

            lon = map_coordinates(self.lon, [y, x],
                                  order=1,
                                  cval=np.nan,
                                  mode='nearest')
            lat = map_coordinates(self.lat, [y, x],
                                  order=1,
                                  cval=np.nan,
                                  mode='nearest')
            return (lon, lat)

    def lonlat2xy(self, lon, lat):
        if self.projected:
            self.__lonlat2xy_parallel__ = False
            return super().lonlat2xy(lon, lat)
        else:
            # For larger arrays, we split and calculate in parallel
            num_elements = len(np.atleast_1d(lon))
            if num_elements > 10000 and not self.__disable_parallel__:
                from multiprocessing import cpu_count
                from concurrent.futures import ThreadPoolExecutor

                self.__lonlat2xy_parallel__ = True

                nproc = cpu_count()
                logger.debug('Running lonlat2xy in parallel using %d threads' %
                             nproc)

                # Chunk arrays
                split_lon = np.array_split(lon, nproc)
                split_lat = np.array_split(lat, nproc)

                with ThreadPoolExecutor() as x:
                    out_x = np.concatenate(
                        list(x.map(self.spl_x, zip(split_lon, split_lat))))
                    out_y = np.concatenate(
                        list(x.map(self.spl_y, zip(split_lon, split_lat))))

                return (out_x, out_y)

            else:
                logger.debug('Calculating lonlat2xy sequentially')
                self.__lonlat2xy_parallel__ = False
                x = self.spl_x(lon, lat)
                y = self.spl_y(lon, lat)
                return (x, y)

    def pixel_size(self):
        if self.projected:
            return super().pixel_size()
        else:
            lons, lats = self.xy2lonlat([self.xmin, self.xmax],
                                        [self.ymin, self.ymin])
            geod = pyproj.Geod(ellps='WGS84')  # Define an ellipsoid
            dist = geod.inv(lons[0], lats[0], lons[1], lats[1],
                            radians=False)[2]
            pixelsize = dist / self.shape[0]
            return pixelsize

    def get_ocean_depth_area_volume(self, lonmin, lonmax, latmin, latmax):
        """Get depth, area and volume of ocean basin within given coordinates"""

        # Extract ocean depth within given boundaries
        background = 'sea_floor_depth_below_sea_level'
        rx, ry = self.lonlat2xy([lonmin, lonmax, lonmax, lonmin], [latmin, latmin, latmax, latmax])
        rx = np.linspace(rx.min(), rx.max(), 10)
        ry = np.linspace(ry.min(), ry.max(), 10)
        data = self.get_variables(background, time=None, x=rx, y=ry)

        x, y = np.meshgrid(data['x'], data['y'])
        lon, lat = self.xy2lonlat(x, y)

        depth = data[background]
        depth = np.ma.masked_where(lon<lonmin, depth)
        depth = np.ma.masked_where(lon>lonmax, depth)
        depth = np.ma.masked_where(lat<latmin, depth)
        depth = np.ma.masked_where(lat>latmax, depth)

        volume = np.nansum(depth*self.pixel_size()*self.pixel_size())
        area = volume/np.nanmean(depth)

        return np.nanmin(depth), np.nanmax(depth), np.nanmean(depth), area, volume

    def _coverage_unit_(self):
        if self.projected:
            return super()._coverage_unit_()
        else:
            return "pixels"

    def _bbox_(self, x, y):
        """
        Find bounding box on grid containing points (x, y)
        """
        ix = (x - self.xmin) / self.delta_x
        ix0, ix1 = np.min(ix), np.max(ix)

        iy = (y - self.ymin) / self.delta_y
        iy0, iy1 = np.min(iy), np.max(iy)

        ix0 = np.max((self.clipped, ix0 - self.buffer)).astype(int)
        iy0 = np.max((self.clipped, iy0 - self.buffer)).astype(int)

        ix1 = np.min((self.numx - self.clipped, ix1 + self.buffer)).astype(int)
        iy1 = np.min((self.numy - self.clipped, iy1 + self.buffer)).astype(int)

        return (ix0, ix1, iy0, iy1)

    def _make_projected_grid_(self, lon, lat, eq_eps=1.e-1):
        """
        Make the projected grid in cases where `lon` and `lat` are present as
        2D variables, but not `x` and `y` and assert that it is approximately
        equidistant.

        Args:

            eq_eps: tolerance for equidistance checks.
        """

        if self.x is not None or self.y is not None:
            logger.error("x and y variables already exist!")

        logger.debug("Finding bounds of reader")
        assert len(lon.shape) == 2
        assert len(lat.shape) == 2

        self.X, self.Y = self.lonlat2xy(lon, lat)
        self.xmin, self.xmax = np.min(self.X[:]), np.max(self.X[:])
        self.ymin, self.ymax = np.min(self.Y[:]), np.max(self.Y[:])

        self.delta_x = np.diff(self.X).flat[0]
        self.delta_y = np.diff(self.Y, axis=0).flat[0]

        self.x = self.X[0, :]
        self.y = self.Y[:, 0]
        self.numx = len(self.x)
        self.numy = len(self.y)

        self.__validate_projected_grid__(eq_eps)

    def __validate_projected_grid__(self, eq_eps=1.e-1):
        """
        Validate that the projected grid is approximately equidistant.

        Args:

            eq_eps: tolerance for equidistance checks.

        Raises:

            AssertionError if not equidistant within `eq_eps`.
        """

        assert np.all(np.abs(self.delta_x - np.diff(self.X)) < eq_eps
                      ), "Grid is not equidistant in X direction"
        assert np.all(np.abs(self.delta_y - np.diff(self.Y, axis=0)) < eq_eps
                      ), "Grid is not equidistant in Y direction"

        assert np.all(
            np.abs(np.tile(self.x, (self.X.shape[0], 1)) - self.X) < eq_eps
        ), "X coordinates are not aligned along Y direction"
        assert np.all(
            np.abs(
                np.tile(np.atleast_2d(self.y).T, (1, self.Y.shape[1])) - self.Y
            ) < eq_eps), "Y coordinates are not aligned along X direction"

    def _slice_variable_(self,
                         var,
                         indxTime=None,
                         indy=None,
                         indx=None,
                         indz=None,
                         indrealization=None):
        """
        Slice variable depending on number of dimensions available.

        Args:

            All arguments can be `slice` objects or index.

        Returns:

            `var` sliced using the slices or indexes necessary to use depending
            on number of dimensions available.

        Raises:

            Unsupported number of dimensions (outside 2..5) raises an exception.

        """
        # NOTE: use match expressions when PEP-634 (Py 3.10) is (widely)
        #       available.
        if var.ndim == 2:
            return var[indy, indx]
        elif var.ndim == 3:
            return var[indxTime, indy, indx]
        elif var.ndim == 4:
            return var[indxTime, indz, indy, indx]
        elif var.ndim == 5:  # Ensemble data
            return var[indxTime, indz, indrealization, indy, indx]
        else:
            raise Exception('Wrong dimension of variable: %s: %d' %
                            (var, var.ndim))



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Gaute Hope, MET Norway

from opendrift.readers.basereader import BaseReader, ContinuousReader

import pyproj
import shapely
import shapely.ops
import shapely.vectorized
import cartopy
import itertools
import logging
logger = logging.getLogger(__name__)

class Reader(BaseReader, ContinuousReader):
    """
    The shape reader can be used to load generic shapes as the 'landmask' variable.

    Args:
        :param shapes: shapely geometries.
        :type shapes: iterable.

        :param proj4_str: Proj.4 string of shape file projection coordinates
                          (default: '+proj=lonlat +ellps=WGS84').
        :type proj4_str: string.
    """
    name = 'shape'
    variables = ['land_binary_mask']
    proj4 = None
    crs   = None
    polys = None
    land  = None
    always_valid = True

    @staticmethod
    def from_shpfiles(shpfiles, proj4_str = '+proj=lonlat +ellps=WGS84', invert=False):
        """
        Construct a shape-reader from shape-files (.shp)

        Args:
            :param shapes: shape-file or files (.shp)
            :type shapes: string or list of file names as strings.

            :param proj4_str: Proj.4 string of shape file projection coordinates
                            (default: '+proj=lonlat +ellps=WGS84').
            :type proj4_str: string.
        """
        if isinstance(shpfiles, list):
            shpfiles = shpfiles
        else:
            shpfiles = [ shpfiles ]

        # reading shapefiles
        shp_iters = []
        for shp in shpfiles:
            logger.debug("Reading shapefile: %s" % shp)
            from cartopy import io
            reader = io.shapereader.Reader(shp)
            shp_iters.append(reader.geometries())

        return Reader(itertools.chain(*shp_iters), proj4_str, invert=invert)

    def __init__(self, shapes, proj4_str = '+proj=lonlat +ellps=WGS84', invert=False):

        self.invert = invert  # True if polygons are lakes and not land areas
        self.proj4 = proj4_str
        self.crs = pyproj.CRS(self.proj4)

        super (Reader, self).__init__ ()

        # Depth
        self.z = None

        # loading and pre-processing shapes
        self.polys = list(shapes) # reads geometries if Shape Readers
        assert len(self.polys) > 0, "no geometries loaded"

        logger.info("Pre-processing %d geometries" % len(self.polys))
        self.land = shapely.ops.unary_union(self.polys)

        self.xmin, self.ymin, self.xmax, self.ymax = self.land.bounds
        self.xmin, self.ymin = self.lonlat2xy(self.xmin, self.ymin)
        self.xmax, self.ymax = self.lonlat2xy(self.xmax, self.ymax)

    def __on_land__(self, x, y):
        if self.invert is False:
            return shapely.vectorized.contains(self.land, x, y)
        else:  # Inverse if polygons are lakes and not land areas
            return 1 - shapely.vectorized.contains(self.land, x, y)

    def get_variables(self, requestedVariables, time = None,
                      x = None, y = None, z = None):
        """
        Get binary mask of whether elements are on land or not.

        Args:
            x (deg[]): longitude (decimal degrees)
            y (deg[]): latitude (decimal degrees)
            ...

        x, y is given in reader local projection.

        Returns:
            Binary mask of point x, y on land.

        """

        self.check_arguments(requestedVariables, time, x, y, z)
        return { 'x' : x, 'y' : y, 'land_binary_mask': self.__on_land__(x,y) }





from abc import abstractmethod
import numpy as np
import scipy
import logging
logger = logging.getLogger(__name__)

from .variables import Variables


class UnstructuredReader(Variables):
    """
    An unstructured reader. Data is gridded irregularily.

    The initial type of grid that this class supports are `triangular prisms
    <https://en.wikipedia.org/wiki/Types_of_mesh#Triangular_prism>`_.
    Unstructured in xy-coordinates, x and y is constant in z. z might be
    non-cartesian (e.g. sigma-levels).

    .. seealso::

        :py:mod:`opendrift.readers`

        :class:`.structured.StructuredReader`

    """
    # Number of parallel threads to use when possible
    PARALLEL_WORKERS = -1

    boundary = None

    # nodes
    x = None
    y = None
    node_variables = None # list of std-name variables defined at nodes
    nodes_idx = None

    # faces
    xc = None
    yc = None
    face_variables = None # list of std-name variables defined at center of faces
    faces_idx = None

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_variables(self, variables, time=None, x=None, y=None, z=None):
        """
        Obtain and return values of the requested variables at all positions
        (x, y, z) closest to given time.

        Returns:

            Dictionary with arrays of length len(time) with values at exact positions x, y and z.
        """

    def _get_variables_interpolated_(self, variables, profiles,
                                   profiles_depth, time,
                                   reader_x, reader_y, z):

        env = self.get_variables(variables, time, reader_x, reader_y, z)

        # We probably have to use an UnstructuredBlock to store closest time-steps, and thus avoid fetching
        # more data on every call.

        logger.debug('Fetched env-before')
        env_profiles = None
        if profiles is not None:
            # Copying data from environment to vertical profiles
            env_profiles = {'z': profiles_depth}
            for var in profiles:
                env_profiles[var] = np.ma.array([env[var], env[var]])

        return env, env_profiles

    def _build_boundary_polygon_(self, x, y):
        """
        Build a polygon of the boundary of the mesh.

        Arguments:
            :param x: Array of node x position, lenght N
            :param y: Array of node y position, length N

        Returns:
            A `shapely.prepareped.prep` `shapely.Polygon`.

            The boundary of the mesh, ideally including holes in the mesh.

        Algorithms:

        .. note::

            Try this alogrithm: https://stackoverflow.com/a/14109211/377927

            Boundary edges (line between two nodes) are only referenced by a single
            triangle.

            1. Find a starting edge segment: [v_start, v_next] (v is vertex or node)

            2. Find another _unvisited_ edge segment [v_i, v_j] that has
               either v_i = v_next or v_j = v_next and add the one not equal to v_next to the polygon.

            3. Reset v_next to the newly added point. Mark edge as visited.

            4. Continue untill we reach v_start.

            The polygon has a rotation, but this should not matter for our purpose
            of checking the bounds.

            Note: In order to find holes in the polygon all points must be scanned.

        Approximate using the convex hull:

        An alternative simple approximation is to use the convex hull of the
        points, but this will miss points along the boundary which form a
        wedge in the boundary (as well as holes in the mesh).

        Holes in the mesh will often be covered by the landmask anyway, so they
        will usually not be a problem.

        """
        from shapely.geometry import Polygon
        from shapely.prepared import prep
        from scipy.spatial import ConvexHull

        P = np.vstack((x, y)).T
        hull = ConvexHull(P)

        boundary = P[hull.vertices, :]
        boundary = prep(Polygon(boundary))

        return boundary

    def covers_positions(self, x, y, z=0):
        """
        Check which points are within boundary of mesh.
        """
        assert self.boundary is not None, "Boundary of mesh has not been prepared by reader"

        # TODO: Check z coordinates
        logger.warning("z-coordinates are not bounds-checked")

        from shapely.vectorized import contains
        return contains(self.boundary, x, y)

    def _build_rtree_(self, x, y):
        """
        Builds an R-tree of x, y
        """
        from rtree import index

        data = ((i, (xy[0], xy[1], xy[0], xy[1]), None) for i, xy in enumerate(zip(x, y)))
        idx = index.Index(data)

        return idx

    def _build_ckdtree_(self, x, y):
        from scipy.spatial import cKDTree
        P = np.vstack((x, y)).T
        return cKDTree(P)

    def __nearest_ckdtree__(self, idx, x, y):
        """
        Return index of nearest point in cKDTree
        """
        q = np.vstack((x, y)).T
        return idx.query(q, k = 1, workers = self.PARALLEL_WORKERS)[1]

    @staticmethod
    def __nearest_rtree__(idx, x, y):
        """
        Take array of points and get nearest point in rtree index.
        """
        return [idx.nearest((x, y, x, y), objects = False) for x, y in zip(x, y)]

    def _nearest_node_(self, x, y):
        """
        Return nearest node (id) for x and y
        """
        return self.__nearest_ckdtree__(self.nodes_idx, x, y)

    def _nearest_face_(self, xc, yc):
        """
        Return nearest element or face (id) for xc and yc
        """
        return self.__nearest_ckdtree__(self.faces_idx, xc, yc)



# -*- coding: utf-8 -*-

"""Vertical structure functions for ROMS

:func:`sdepth`
  Depth of s-levels
:func:`zslice`
  Slice a 3D field in s-coordinates to fixed depth
:func:`multi_zslice`
  Slice a 3D field to several depth levels
:func:`z_average`
  Vertical average of a 3D field
:func:`s_stretch`
  Compute vertical stretching arrays Cs_r or Cs_w

"""

# -----------------------------------
# Bjørn Ådlandsvik <bjorn@imr.no>
# Institute of Marine Research
# Bergen, Norway
# 2010-09-30
# -----------------------------------

import numpy as np

def sdepth(H, Hc, C, stagger="rho", Vtransform=1):
    """Depth of s-levels

    *H* : arraylike
      Bottom depths [meter, positive]

    *Hc* : scalar
       Critical depth

    *cs_r* : 1D array
       s-level stretching curve

    *stagger* : [ 'rho' | 'w' ]

    *Vtransform* : [ 1 | 2 ]
       defines the transform used, defaults 1 = Song-Haidvogel

    Returns an array with ndim = H.ndim + 1 and
    shape = cs_r.shape + H.shape with the depths of the
    mid-points in the s-levels.

    Typical usage::

    >>> fid = Dataset(roms_file)
    >>> H = fid.variables['h'][:, :]
    >>> C = fid.variables['Cs_r'][:]
    >>> Hc = fid.variables['hc'].getValue()
    >>> z_rho = sdepth(H, Hc, C)

    """
    H = np.asarray(H)
    Hshape = H.shape      # Save the shape of H
    H = H.ravel()         # and make H 1D for easy shape maniplation
    C = np.asarray(C)
    N = len(C)
    outshape = (N,) + Hshape       # Shape of output
    if stagger == 'rho':
        S = -1.0 + (0.5+np.arange(N))/N    # Unstretched coordinates
    elif stagger == 'w':
        S = np.linspace(-1.0, 0.0, N)
    else:
        raise ValueError("stagger must be 'rho' or 'w'")

    if Vtransform == 1:         # Default transform by Song and Haidvogel
        A = Hc * (S - C)[:, None]
        B = np.outer(C, H)
        return (A + B).reshape(outshape)

    elif Vtransform == 2:       # New transform by Shchepetkin
        N = Hc*S[:, None] + np.outer(C, H)
        D = (1.0 + Hc/H)
        return (N/D).reshape(outshape)

    else:
        raise ValueError("Unknown Vtransform")

# ------------------------------------


def sdepth_w(H, Hc, cs_w):
    """Return depth of w-points in s-levels

    Kept for backwards compatibility
    use *sdepth(H, Hc, cs_w, stagger='w')* instead

    """
    return sdepth(H, Hc, cs_w, stagger='w')

# ------------------------------------------
# Vertical slicing e.t.c.
# ------------------------------------------


def zslice(F, S, z):
    """Vertical slice of a 3D ROMS field

    Vertical interpolation of a field in s-coordinates to
    (possibly varying) depth level

    *F* : array with vertical profiles, first dimension is vertical

    *S* : array with depths of the F-values,

    *z* : Depth level(s) for output, scalar or ``shape = F.shape[1:]``
          The z values should be negative

    Return value : array, `shape = F.shape[1:]`, the vertical slice

    Example:
    H is an array of depths (positive values)
    Hc is the critical depth
    C is 1D containing the s-coordinate stretching at rho-points
    returns F50, interpolated values at 50 meter with F50.shape = H.shape

    >>> z_rho = sdepth(H, Hc, C)
    >>> F50 = zslice(F, z_rho, -50.0)

    """

    # TODO:
    # Option to Save A, D, Dm
    #   => faster interpolate more fields to same depth

    F = np.asarray(F)
    S = np.asarray(S)
    z = np.asarray(z, dtype='float')
    Fshape = F.shape  # Save original shape
    if S.shape != Fshape:
        raise ValueError("F and z_r must have same shape")
    if z.shape and z.shape != Fshape[1:]:
        raise ValueError("z must be scalar or have shape = F.shape[1:]")

    # Flatten all non-vertical dimensions
    N = F.shape[0]        # Length of vertical dimension
    M = F.size // N        # Combined length of horizontal dimension(s)
    F = F.reshape((N, M))
    S = S.reshape((N, M))
    if z.shape:
        z = z.reshape((M,))

    # Find integer array C with shape (M,)
    # with S[C[i]-1, i] < z <= S[C[i], i]
    # C = np.apply_along_axis(np.searchsorted, 0, S, z)
    # but the following is much faster
    C = np.sum(S < z, axis=0)
    C = C.clip(1, N-1)

    # For vectorisation
    # construct index array tuples D and Dm such that
    #   F[D][i]  = F[C[i], i]
    #   F[Dm][i] = F[C[i]-1, i]
    I = np.arange(M, dtype='int')
    D = (C, I)
    Dm = (C-1, I)

    # Compute interpolation weights
    A = (z - S[Dm]) / (S[D]-S[Dm])
    A = A.clip(0.0, 1.0)   # Control the extrapolation

    # Do the linear interpolation
    R = (1-A)*F[Dm]+A*F[D]

    # Give the result the correct s
    R = R.reshape(Fshape[1:])

    return R

# -----------------------------------------------


def multi_zslice(F, S, Z):

    """Slice a 3D ROMS field to fixed depth

    Vertical interpolation of a field in s-coordinates to
    fixed vertical level

    *F* : array of with vertical profiles, first dimension is vertical

    *S* : array with depth of s-levels (at rho-points)
        1D (constant depth) or  S.shape = F.shape

    *Z* : single depth value, negative

    Returns : array, ``shape = F.shape[1:]`` the vertical slice

    """

    # TODO:
    # Option to Save A, D, Dm
    #   => faster interpolate more fields to same depth

    F = np.asarray(F)
    S = np.asarray(S)
    Fshape = F.shape  # Save original shape

    # Flat all dimensions after first
    N = F.shape[0]
    M = F.size // N
    F = F.reshape((N, M))
    S = S.reshape((N, M))

    # Make z.shape = (M,)
    Z = np.asarray(Z, dtype='float')
    # Valid possibilities
    # 1) Z = single scalar (shape = ()), one constant value
    # 2) Z = 1D array, shape=(kmax), a set of constant depths
    # 3) Z = 2D or more, reshapeable to (kmax, M)

    if Z.ndim == 0:
        Z = Z + np.zeros((1, M))
        kmax = 1
    elif Z.ndim == 1:
        kmax = Z.size
        Z = Z[:, np.newaxis] + np.zeros((kmax, M))
    else:
        kmax = Z.size // M
        Z = Z.reshape((kmax, M))

    # Find C, C.shape = (kmax, M) such that
    # z_r[C[k,i]-1, i] < Z[k] <= z_r[C[k,i], i]

    # shape: kmax, N, M => kmax, M
    C = np.sum(S[np.newaxis, :, :] < Z[:, np.newaxis, :], axis=1)
    C = C.clip(1, N-1)

    # Horizontal index
    I = np.arange(M, dtype=int)

    # Compute interpolation weights
    A = (Z - S[(C-1, I)])/(S[(C, I)]-S[(C-1, I)])
    A = A.clip(0.0, 1.0)   # Control the extrapolation

    # Do the interpolation
    R = (1-A)*F[(C-1, I)]+A*F[(C, I)]

    # Give the result the correct shape
    R = R.reshape((kmax,) + Fshape[1:])

    return R, (A, C, I, kmax)

# ------------------------------------------------------


def z_average(F, z_r, z0, z1):
    """Slice a 3D ROMS field to fixed depth

    Vertical interpolation of a field in s-coordinates to
    fixed vertical level

    *F* : array
      Vertical profiles, first dimension is vertical
    *z_r* : array
      Depth of s-levels (at rho-points), requires `z_r.shape = F.shape`
    *z0*, *z1* : floats
      Single depth values with z0 <= z1 <= 0

    return value : array
      `shape = F.shape[1:]`, the vertical average

    """

    F = np.asarray(F)
    z_r = np.asarray(z_r)
    Fshape = F.shape  # Save original shape

    # Flatten all dimensions after first
    N = F.shape[0]
    M = F.size // N
    F = F.reshape((N, M))
    z_r = z_r.reshape((N, M))

    # z0, z1 are scalars or horizontal arrays
    z0 = np.asarray(z0)
    if z0.shape:  # Array, must be 2D
        z0 = z0.reshape((M,))
    z1 = np.asarray(z1)
    if z1.shape:
        z1 = z1.reshape((M,))

    # Bracket z0, i.e.
    # Find integer array C0 with shape (M,)
    # with z_r[C0[i]-1, i] < z0 <= z_r[C0[i], i]
    # Can be done with:
    #   C0 = np.apply_along_axis(np.searchsorted, 0, z_r, z0)
    # but the following is much faster
    C0 = np.sum(z_r < z0, axis=0)
    C0 = C0.clip(1, N-1)       # Clip to avoid illegal indices

    # Bracket z1
    C1 = np.sum(z_r < z1, axis=0)
    C1 = C1.clip(1, N-1)

    # Use advanced indexing for vectorisation
    #   F[(C0,I)][i]  = F[C0[i], i]
    I = np.arange(M, dtype='int')
    # Interpolate F to the two levels
    A0 = (z0 - z_r[(C0-1, I)]) / (z_r[(C0, I)]-z_r[(C0-1, I)])
    A0 = A0.clip(0.0, 1.0)   # Control the extrapolation
    F0 = (1-A0)*F[(C0-1, I)]+A0*F[(C0, I)]
    A1 = (z1 - z_r[(C1-1, I)])/(z_r[(C1, I)]-z_r[(C1-1, I)])
    A1 = A1.clip(0.0, 1.0)
    F1 = (1-A1)*F[(C1-1, I)] + A1*F[(C1, I)]

    # Find indices again (unclipped)
    C0 = np.sum(z_r < z0, axis=0)
    C1 = np.sum(z_r < z1, axis=0)

    R = np.zeros(M, dtype=np.float64)
    X = np.zeros(N+2, dtype=np.float64)
    Y = np.zeros(N+2, dtype=np.float64)
    z0 = z0 + R    # Make sure they are spatial arrays
    z1 = z1 + R    # For indexing below

    for i in I:
        X[:] = 0.0
        Y[:] = 0.0
        nz = C1[i] - C0[i]  # Number of rho-points between z0 and z1
        # Set up arrays for trapezoidal integration
        X[0] = z0[i]
        X[1:nz+1] = z_r[C0[i]:C1[i], i]
        X[nz+1] = z1[i]
        Y[0] = F0[i]
        Y[1:nz+1] = F[C0[i]:C1[i], i]
        Y[nz+1] = F1[i]
        # Perform the integration
        R[i] = 0.5 * np.dot(X[1:nz+2]-X[0:nz+1], Y[1:nz+2]+Y[0:nz+1])

    # Compute average and revert to correct shape
    R = R / (z1-z0)
    R = R.reshape(Fshape[1:])

    return R

# ----------------------------------


def s_stretch(N, theta_s, theta_b, stagger='rho', Vstretching=1):
    """Compute a s-level stretching array

    *N* : Number of vertical levels

    *theta_s* : Surface stretching factor

    *theta_b* : Bottom stretching factor

    *stagger* : "rho"|"w"

    *Vstretching* : 1|2|4

    """

    if stagger == 'rho':
        S = -1.0 + (0.5+np.arange(N))/N
    elif stagger == "w":
        S = np.linspace(-1.0, 0.0, N+1)
    else:
        raise ValueError("stagger must be 'rho' or 'w'")

    if Vstretching == 1:
        cff1 = 1.0 / np.sinh(theta_s)
        cff2 = 0.5 / np.tanh(0.5*theta_s)
        return ((1.0-theta_b)*cff1*np.sinh(theta_s*S)
                + theta_b*(cff2*np.tanh(theta_s*(S+0.5))-0.5))

    elif Vstretching == 2:
        a, b = 1.0, 1.0
        Csur = (1 - np.cosh(theta_s * S))/(np.cosh(theta_s) - 1)
        Cbot = np.sinh(theta_b * (S+1)) / np.sinh(theta_b) - 1
        mu = (S+1)**a * (1 + (a/b)*(1-(S+1)**b))
        return mu*Csur + (1-mu)*Cbot

    elif Vstretching == 4:
        C = (1 - np.cosh(theta_s * S)) / (np.cosh(theta_s) - 1)
        C = (np.exp(theta_b * C) - 1) / (1 - np.exp(-theta_b))
        return C

    else:
        raise ValueError("Unknown Vstretching")


# wrapper for backwards compatibility
def s_stretch_w(N, theta_s, theta_b, Vstretching=1):
    """Obsolete use *s_stretch* instead"""

    return s_stretch(N, theta_s, theta_b,
                     stagger='w', Vstretching=Vstretching)



from .label import Label
from .synonym import Synonym
from .sara import SARAFraction, SARADensity, MolecularWeight
# from .measurement import (ProductType,
#                           MeasurementBase,
#                           Temperature,
#                           Length,
#                           Mass,
#                           MassFraction,
#                           VolumeFraction,
#                           Density,
#                           DynamicViscosity,
#                           KinematicViscosity,
#                           Pressure,
#                           NeedleAdhesion,
#                           InterfacialTension,
#                           Unitless,
#                           Dimensionless)
